package cds.astro;

/*==================================================================
                Otype class (includes also Odef)
 *==================================================================*/

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*; // for parseException

// import cds.util.*;
// import cds.util.net.UFile;

/**
 * The <em>Otype</em> is a classification of the astronomical objects
 * in a hierarchical structure.
 * 
 * @author Francois Ochsenbein<br>
 * @author Sylvaine Jaehn (Stage UTBM)<br>
 * @author Benoit Baranne (Stagiaire UTBM)<br>
 * @author Marc Wenger/CDS<br>
 * @author Anais Oberto/CDS<br>
 * @version 1.0 Sep 2002<br>
 * @version 1.1 Oct 2002 SJ ajout de la methode numericToHexa()<br>
 * @version 1.2 Nov 2002 SJ ajout du test : <br>
 *                          if (!initialized) init(); ds constructeur otype sans parametre<br>
 *                          ajout de la methode hexaToNumeric(String strHex)<br>
 * @version 1.3 Jan 2003 SJ methode init() mise public<br>
 * @version 1.4 jan 2004 BB ajout de la methode equals.<br>
 * @version 1.5 Dec 2004 MW Access to the different levels. getValue(level)<br>
 *                          Add new constructors<br>
 * @version 1.6 May 2005 MW Use UFile for accessing the description file<br>
 * @version 1.7 May 2005 MW add OFormatable interface and its method toString(String)<br>
 * @version 1.8 Sep 2005 MW add methods needed for sql requests: getQueryMask()<br>
 * @version 1.9 May 2006 AO add method isUnknown()<br>
 * @version 1.10 May 2006 MW Implement the Data interface<br>
 * @version 1.11 May 2006 MW add getCompatibleTypes() needed for database requests.<br>
 *                           fix a bug: keep lines larger than 8 (not 10) chars (see '  CV* / X')
 * @version 1.12 Sep 2006 MW Override the hashCode() method
 * @version 1.13 Sep 2006 MW Add a toString(char) method
 **/

public class Otype implements Serializable, Cloneable { // , Data, ValueList, OFormatable 

    private static final int WARNING_1 = 1;
    private static final int WARNING_2 = 2;
    private static final int WARNING_3 = 3;
    private static final int WARNING_4 = 4;
	// The Otype is just a number -- the bits are arranged to map to
    // the hierarchical structure.
    protected int otype; // The Otype is just a number!!
    public static final int UNKNOWN = 0;
    private static final int NODATA = 0;
    private static final int VARIABLE = 9; 
    private static final int MULTIPLE = 12;
    private static final int DOUBLE = 13;
    private static final int STAR = 14;
    
    
    // Variables common to the Class: all Definitions
    static public boolean DEBUG = false;
    static boolean initialized = false;
    // in case of structure modification, change the getValue() method
    // if changing the field size, keep the rightmost bit unused (see shift 1 in merge())
    static private final int B0 = 4; // Wavelength
    static private final int B1 = 4; // Main Class
    static private final int B2 = 5; // Subclass
    static private final int B3 = 4; // Nova/SN
    static private final int Bf = 15; // Not Yet Used
    static private final char[] dig = "0123456789".toCharArray();
    static private final int[] bits = {
    	B0, B1, B2, B3, Bf};
    static private final int[] Nt = {
    	(1 << B0) - 1, 
    	(1 << B1) - 1, 
    	(1 << B2) - 1, 
    	(1 << B3) - 1, 
    	(1 << Bf) - 1};
    static protected final int[] Nlev = new int[bits.length + 1];
    //  0, 0xf0000000, 0xff000000,		// Levels 0 1 2
    //  0xfff80000, 0xffff8000, 0xffffffff	// Levels 3 4 5
    static private HashMap symbols = new HashMap(300);
    static protected ArrayList otypes = new ArrayList(200);
    static protected ArrayList otrees = new ArrayList(100);
    // static private final String def_url = 
    // 	"http://simbad.u-strasbg.fr/otype.def";
    static private final String def_url =
    	"http://localhost:2001/Tables/Otype/otype.def";
    
    static {
    	// Nlev table contains truncation masks:
    	// [0] = 0x00000000
    	// [1] = 0xF0000000
    	// [2] = 0xFF000000
    	// [3] = 0xFFF80000
    	// [4] = 0xFFFF8000
    	// [5] = 0xFFFFFFFF
    	
    	int i, o;
    	o = 0xffffffff;
    	i = bits.length;
    	while (i > 0) {
    		Nlev[i] = o;
    		o <<= bits[--i];
    	}
    	Nlev[0] = 0;
    }

    //==================================================================
    //			Constructors
    //==================================================================
    
    /** Define the default Otype (object of unknown nature)
     */
    public Otype() {
    	if (!initialized) {
    		init();
    	}
    	otype = NODATA;
    }
    
    /** Define the Otype from a String
     * @param text the Otype as a string
     * @throws ParseException when otype is not understandable
     */
    public Otype(String text) throws ParseException {
    	set(text);
    }
    
    /** Define the Otype from an otype code (integer)
     * @param otype the Otype as a coded integer
     * @throws ParseException when otype is not understandable
     */
    public Otype(int otype) throws ParseException {
    	this.otype = otype;
    	verify();
    }
    
    /**
     * Copy constructor
     * @param ot Otype object to copy
     */
    public Otype(Otype ot) {
        this();
        otype = ot.otype;
    }
    
    /**
     * Check if this type has a value. An unknown type is different from nodata.
     * @return true if it has a value, false otherwise.
     */
    public boolean hasValue() {
    	return ! (otype == NODATA); //NoData.isNodata(otype);
    }
    
    /**
     * Reset this object type to its NODATA value. The NODATA value is
     * different from the type 'unknown'.
     */
    public void reset() {
    	otype = NODATA;
    }
    
    /**
     * Get the binary type
     * @return the integer value
     */
    public final int getBinType() {
    	return otype;
    }

    //==================================================================
    // Upload the list of Object Types
    //==================================================================
    /** 
     * Read the List of the Object Types
     * @param	plug the input stream
     **/
    private static void upload(BufferedReader plug) {
    	//File odef = new File("otype.def");
    	String s1, s2, s3;
    	Odef odef; // An Object Type
    	char[] chline;
    	String line;
    	int i, k;
    	int bintype;
    	bintype = 0xffffffff;
    	i = bits.length;
    	while (i > 0) {
    		Nlev[i] = bintype;
    		bintype <<= bits[--i];
    	}
    	Nlev[0] = 0;
    	while (true) {
    		try {
    			line = plug.readLine();
    		}
    		catch (Exception e) {
    			System.err.println(e);
    			return;
    		}
    		if (line == null) {
    			break;
    		}
    		if (line.length() < 8) {
    			continue;
    		}
    		chline = line.toCharArray();
    		if (chline[0] == '#') {
    			continue; // comment
    		}
    		if (chline[0] == '-') {
    			continue; // ignored
    		}
    		if (chline[0] == '(') {
    			continue; // ignored
    		}
    		if (chline[0] == '+') {
    			continue; // ignored
    		}
    		for (i=0; (i < chline.length) && Character.isWhitespace(chline[i]); i++);
    		if (i >= chline.length) {
    			continue;
    		}
    		//if (DEBUG) System.out.println("Got: " + line);
    		// Find the ':' which defines Synonyms
    		k = i;
    		while ( (i < chline.length) && (chline[i] != ':') && (chline[i] != '/')) {
    			i++;
    			/* Interpret the line: numeric type / equivalences */
    		}
    		if (Character.isDigit(chline[k])) {
    		    /* Here the line is like
    		      00.00.00.0: Unknown     ?    "Object of unknown nature"
    		    --------------------------------------------------------*/
    			bintype = ntype(new String(chline, k, i - k));
    			if (i < chline.length) {
    				i++; // Skip the :
    			}
    			while ( (i < chline.length) && Character.isWhitespace(chline[i])) {
    				i++;
    			}
    			k = i;
    			while ( (i < chline.length) && (!Character.isWhitespace(chline[i]))) {
    				i++;
    			}
    			s1 = new String(chline, k, i - k);
    			while ( (i < chline.length) && Character.isWhitespace(chline[i])) {
    				i++;
    			}
    			if ( (i < chline.length) && (chline[i] != '"')) { // "
    				k = i;
    				while ( (i < chline.length) && (!Character.isWhitespace(chline[i]))) {
    					i++;
    				}
    				s2 = new String(chline, k, i - k);
    				while ( (i < chline.length) && Character.isWhitespace(chline[i])) {
    					i++;
    				}
    			}
    			else {
    				s2 = s1; // The 3-letter abbreviation
    			}
    			if ( (i < chline.length) && (chline[i] == '"')) { // "
    				k = ++i;
    				while ( (i < chline.length) && (chline[i] != '"')) {
    					i++; // "
    				}
    				s3 = new String(chline, k, i - k);
    				if (i < chline.length) {
    					i++;
    				}
    				while ( (i < chline.length) && Character.isWhitespace(chline[i])) {
    					i++;
    				}
    			}
    			else {
    				s3 = s2;
    			}
    			odef = new Odef(bintype, s1, s2, s3);
    			otypes.add(odef);
    			hadd(s1, odef);
    			if (s2 != s1) {
    				hadd(s2.trim(), odef);
    			}
    			if (s3 != s2) {
    				hadd(s3, odef);
    				/* Are there other synonyms ? */
    			}
    			while (i < chline.length) {
    				k = i;
    				while ( (i < chline.length) && (!Character.isWhitespace(chline[i]))) {
    					i++;
    				}
    				s3 = new String(chline, k, i - k);
    				hadd(s3, odef);
    				while ( (i < chline.length) && Character.isWhitespace(chline[i])) {
    					i++;
    				}
    			}
    			continue;
    		}
    		if (i >= chline.length) {
    			continue; // Empty Line
    		}
    		s1 = new String(chline, k, i - k);
    		s1 = s1.trim();
    		k = i + 1; // Second part
    		s2 = new String(chline, k, chline.length - k);
    		s2 = s2.trim();
    		//if(DEBUG) System.out.println("....additional: " + s1 + a[i] + s2);
    		if (chline[i] == ':') { // There is a : ==> Synonym
    			hadd(s1, (Odef) symbols.get(s2.toLowerCase()));
    			continue;
    		}
    		if (chline[i] == '/') { // There is a / ==> Subtree
    			otrees.add(symbols.get(s1.toLowerCase()));
    			otrees.add(symbols.get(s2.toLowerCase()));
    		}
    	}
    	initialized = true;
    	otypes.trimToSize();
    	otrees.trimToSize();
    	/*if (DEBUG) System.out.println("....Otype init: " + otypes.size()
    	  + "otypes, " + otrees.size() + "subtrees, " + Hsymbols.size()
    	  + "symbols");*/
    }

    /** 
     * Insert a definition -- verify it's not alreay in
     * @param	symb the symbol of this object type
     * @param	odef the object type definition
     **/
    private static void hadd(String symb, Odef odef) {
    	String s = symb.toLowerCase();
    	Object o = symbols.put(s, odef);
    	if (o != null) {
    		if (odef.equals(o)) {
    			return;
    		}
    		System.err.println("****Otype: '" + symb + "' ambiguous: "
				 + odef + " <> " + o);
    	}
    }
    
    //==================================================================
    // 		Initialisations
    //==================================================================

    /** Initialisation
     */
    public static void init() {
    	DataInputStream plug;
    	URL odef;
//    	System.out.println("---> Otype.init() sans parametres appele");

    	try {
    		odef = new URL(def_url);
    		plug = new DataInputStream(odef.openStream());
    		upload(new BufferedReader(new InputStreamReader(plug)));
    	}
    	catch (Exception e) {
    		System.err.println(e);
    	}
    }

    /** Initialisation
     * @param	filename name of file with the list of Object Types
     * @throws Exception e.g. IOException if file not existing, parsing...
     */
    public static void init(String filename) throws Exception {
//    	UFile ufile = new UFile();
//    	ufile.getFile(filename);
//    	upload(new BufferedReader(ufile.getReader()));
    	File odef;
    	FileInputStream plug;
    	odef = new File(filename);
    	plug = new FileInputStream(odef);
    	upload(new BufferedReader(new InputStreamReader(plug)));
    }

    /** Compute the level in the hierarchy
     * @param	otype an object type
     * @return	the level (0 1 2 3 4)
     **/
    protected static final int level(int otype) {
    	int i;
    	for (i = 0; otype != 0; otype <<= bits[i++]) ;
    	return (i);
    }

    /** 
     * Test whether ot is in class cl
     * @param	cl a 'class'
     * @param	ot a 'subclass'
     * @return	true if ot if subclass of cl
     **/
    private static final boolean otinclass(int cl, int ot) {
    	int i = level(cl);
    	return ( (ot & Nlev[i]) == cl);
    }
    
    //==================================================================
    //		Internal Routines
    //==================================================================

    /** 
     * Edit the otype as xx.xx.xx.xx
     * @param	buf the buffer to which the ascii equivalent of otype
     *			is appended.
     * @param	otype the object type
     **/
    protected static void edit(StringBuffer buf, int otype) {
    	int[] otc = new int[4];
    	int o = otype >>> Bf;
    	int i;
    	for (i = bits.length - 1; --i >= 0; o >>>= bits[i]) {
    		otc[i] = o & Nt[i];
    	}
    	for (i = 0; i < 4; i++) {
    		buf.append(dig[otc[i] / 10]);
    		buf.append(dig[otc[i] % 10]);
    		buf.append('.');
    	}
    	i = buf.length();
    	buf.setLength(i - 1);
    }

    /** Interpret the 4 numbers as an Otype class.
     * @param	s the text to interpret
     * @return	The object type / -1 when error
     **/
    private static final int ntype(String s) {
    	char a[] = s.toCharArray();
    	int ot = 0;
    	int i, h, n;
    	for (i = 0, h = 0; (i < a.length) && (h < bits.length); h++) {
    		ot <<= bits[h];
    		if (a[i] == '.') {
    			i++;
    		}
    		for (n = 0; (i < a.length) && Character.isDigit(a[i]); i++) {
    			n = (n * 10) + Character.digit(a[i], 10);
    		}
    		if (n >= (1 << bits[h])) {
    			return ( -1);
    		}
    		ot |= n;
    	}
    	while (h < bits.length) {
    		ot <<= bits[h++];
    	}
    	return (ot);
    }

    /** Verify the number corresponds to an existing Otype
     * @throws ParseException when the otype is incorrect
     */
    private final void verify() throws ParseException {
    	if (! this.hasValue()) return;
    	Odef o = new Odef(otype, null, null, null);
    	if (otypes.indexOf(o) < 0) {
    		throw new ParseException
    		("****Otype: invalid number to assignation: " + otype, 0);
    	}
    }

    //==================================================================
    // Manipulation of Otypes
    //==================================================================

    /**
     * Get the value of a particular level (0-3)
     * @param level from 0 (main level) to 3 (today lowest one)
     * @return the level value
     */
    public int getValue(int level) {
    	if (hasValue()) {
	    	switch(level) {
	    		case 0: return (otype >>> (B1+B2+B3+Bf)) & Nt[level];
	    		case 1: return (otype >>> (B2+B3+Bf)) & Nt[level];
	    		case 2: return (otype >>> (B3+Bf)) & Nt[level];
	    		case 3: return (otype >>> Bf) & Nt[level];
	    		default:
	    	}
    	}
    	return 0;
    }
    
    /** Truncate the otype to some upper level (e.g. Algol --> V*)
     * @param	lev the level (between 1 and 4)
     **/
    public final void truncate(int lev) {
    	if (! hasValue()) return;
    	
    	if (!initialized) {
    		init();
    	}
    	if (lev < 0) {
    		lev = 0;
    	}
    	if (lev > bits.length) {
    		lev = bits.length;
    	}
    	otype &= Nlev[lev];
    }

    /** Check if is a subtype (in same branch)
     * @param	t1 the other Otype
     * @return	true t1 is an upper class
     **/
    public final boolean implies(Otype t1) {
    	if (hasValue() && t1.hasValue())
    		return (implies(otype, t1.otype));
    	else
    		return false;
    }

    /**
     * Checks whether t0 implies t1
     * @param t0 first otype
     * @param t1 second otype
     * @return true if t0 implies t1
     */
    private static final boolean implies(int t0, int t1) {
    	int lev;
    	if (!initialized) {
    		init();
    	}
    	lev = level(t1);
    	return ( (t0 & Nlev[lev]) == t1);
    }

    /** Check if the 2 Otypes are compatible
     * @param	ot1 the other Otype
     * @return	true t1 is compatible
     **/
    public final boolean agrees(Otype ot1) {
    	Otype type, clas;
    	int i;
    	Odef o;
    	if (ot1.otype == otype) {
    		return (true);
    	}
    	// Order in clas (less accurate) / type (more accurate)
    	type = merge(ot1);
    	clas = ot1 == type ? this : ot1;
    	if (type.implies(clas)) {
    		return (true);
    	}
    	for (i = 0; i < otrees.size(); i += 2) {
    		o = (Odef) otrees.get(i);
    		if (!implies(type.otype, o.otype)) {
    			continue;
    		}
    		o = (Odef) otrees.get(i + 1);
    		if (implies(o.otype, clas.otype)) {
    			return (true);
    		}
    	}
    	return (false);
    }
    
    /**
     * For the current type, get all compatible types in other branches
     * @return an array list of all compatible types. The first item is the current type
     */
    public ArrayList getCompatibleTypes() {
    	ArrayList ctypes = new ArrayList();
    	ctypes.add(this);
    	int omask = Nlev[level(otype)];
    	
	    int inc = 1; // Index in otrees
    	while (inc < otrees.size()) {
    		Odef odef = (Odef) otrees.get(inc);
    		if ( (odef.otype & omask) == otype) {
    			odef = (Odef) otrees.get(inc - 1);
    			try {
					ctypes.add(new Otype(odef.otype));
				} catch (ParseException e) {
					// should never happen, so do nothing
				}
    		}
    		inc += 2;
    	}
    	
    	return ctypes;
    }

    /** Choose the most accurate Otype: the deepest one ine the hierarchy.
     * @param	ot1 a second otype
     * @return	the 'most accurate' otype
     **/
    public final Otype merge(Otype ot1) {
    	int lev, lev1;
    	// elimnate the cases where one otype has no value
    	if (! this.hasValue()) return ot1;
    	if (! ot1.hasValue()) return this;
    	
    	if (!initialized) {
    		init();
    	}
    	lev = level(otype);
    	lev1 = level(ot1.otype);
    	/* The otype should be UNSIGNED --- but unsigned int in not
    	 known in Java. Therefore compare the right-shifted values
    	 */
    	if (lev == lev1) {
    		return
    			( (otype >>> 1) > (ot1.otype >>> 1) ? this : ot1);
    	}
    	if (lev < lev1) {
    		return (ot1);
    	}
    	return (this);
    }

    public final Otype mergepl(Otype ot2) throws ParseException {
    	
    	int i1,i2,i11,i22,l1,l2;
    	if (!initialized) { init() ; }
    	
    	StringBuffer numotype1 = new StringBuffer();
    	StringBuffer numotype2 = new StringBuffer();
    	
    	edit(numotype1,otype) ;
    	edit(numotype2,ot2.otype) ;
    	
    	StringTokenizer st1 = new StringTokenizer(numotype1.toString(), ".");
    	StringTokenizer st2 = new StringTokenizer(numotype2.toString(), ".");
    	
    	/* first level */
    	i1 = Integer.parseInt(st1.nextToken());
    	/* second level */ 
    	i11 = Integer.parseInt(st1.nextToken()); 
    	/* first level */
    	i2 = Integer.parseInt(st2.nextToken()); 
    	/* second level */
    	i22 = Integer.parseInt(st2.nextToken());
    	
    	/* Types d'objets composites distincts on ne change rien */
    	if ((i1 == MULTIPLE) && (i2 == MULTIPLE)) 
    		throw new ParseException("Warning : Types d'objets composites differents", WARNING_3) ;
    	
    	/* si les niveaux 1 sont identiques et dif de MULTIPLE on retourne le plus grand des deux sous-types */
    	if (i1 == i2){ return ( (otype >>> 1) > (ot2.otype >>> 1) ? this : ot2) ; }
    	
    	/* niveaux 1 distincts */
    	    	
    	/* si objets divers de niveau 1 distinct on ne change rien */
    	if((i1<=7)&&(i2 <=7)) 
    		return this ;
    	
    	/* hors objets divers */
    	else if ((i1 >= VARIABLE)&&(i2 >=VARIABLE))
    	{
    		/* etoiles variables et binaires: on choisit le plus precis et si
    		 aussi precis on laisse inchang */
    		if(((i1 == MULTIPLE) && (i2 == STAR)&& (i11 == DOUBLE) && (i22 == VARIABLE)) ||
    				((i2 == MULTIPLE) && (i1 == STAR)&& (i22 == DOUBLE) && (i11 == VARIABLE)))
    		{
    			l1 = level(otype) ;
    			l2 = level(ot2.otype);
    			if (l1 > l2) return this ;
    			if (l2 > l1) return ot2 ;
    			if(l1 == l2) 
    				throw new ParseException("Warning : Etoile variable et binaire", WARNING_4) ;
    		}
    		/* dans les autres cas on dit que les types d'objet sont incompatibles 
    		 et on laisse inchang*/
    		else throw new ParseException("Warning : Types d'objets merges incompatibles",WARNING_2) ;
    	}
    	/* autres cas on prend le plus grand des deux types */
    	else { return ( (otype >>> 1) > (ot2.otype >>> 1) ? this : ot2) ; }
    	
    	return this ;
    	
    }
    

          

    //==================================================================
    // 		Setting an Otype from input details
    //==================================================================

    /** Convert an integer into an Otype
     * @param	o a number representing an otype
     * @throws ParseException when the text can' be interpreted.
     **/
    protected final void set(int o) throws ParseException {
    	if (!initialized) {
    		init();
    	}
    	otype = o;
    	verify();
    }

    /** Convert a text into an object type
     * @param	text the text representing the otype, either in numeric,
     *		or as one of the available abbreviations
     * @throws ParseException when the text can' be interpreted.
     **/
    public final void set(String text) throws ParseException {
    	Odef odef;
    	// check first if the otype is NODATA
    	if (text.length() == 0) { // NoData.parseNodata(text) != 0) 
    		otype = NODATA;
    		return;
    	}
    	if (!initialized) {
    		init();
    	}
    	odef = (Odef) symbols.get(text.toLowerCase());
    	if (odef != null) {
    		otype = odef.otype;
    	}
    	else if (text.length() < 1) {
    		otype = 0;
    	}
    	else if (Character.isDigit(text.charAt(0))) {
    		otype = ntype(text);
    		verify();
    	}
    	else {
    		throw new ParseException("****Otype: " + text, 0);
    	}
    }

    //==================================================================
    //	Give the list of Otypes
    //==================================================================

    /** List in an enumeration all Otypes belonging to a class.
     * @return	An enumeration of Otypes in subtree
     **/
    public Enumeration list() {
    	return new Enumeration() {
    		int nbmoc = 0;
    	    int o = otype; // Current otype
    	    int mask = 0;
    	    int pos = -1; // Index in otypes
    	    int ori = otype; // Originally asked type
    	    int omask = 0; // Mask corresponding to original type
            int inc = 1; // Index in otrees
    	    public boolean hasMoreElements() {
    	    	Odef odef;
    	    	if(DEBUG) {
		    StringBuffer buf = new 
			StringBuffer("....hasMore(class=");
		    edit(buf, o);
		    buf.append(") pos=" + pos + ", inc=" + inc);
		    System.out.println(buf);
		}
    	    	if (pos < 0) {
    	    		if (!initialized) {
    	    			init();
    	    		}
    	    		odef = new Odef(o, null, null, null);
    	    		pos = otypes.indexOf(odef);
    	    		omask = mask = Nlev[level(o)];
    	    		if (DEBUG) System.out.println("init : odef="+odef.abbr+" pos="+pos+
    	    				" omask="+Integer.toHexString(omask)+" mask="+Integer.toHexString(mask));
    	    		return (pos >= 0);
    	    	}
    	    	if (DEBUG) System.out.println("debut: pos="+pos+
 	    				" omask="+Integer.toHexString(omask)+" mask="+Integer.toHexString(mask));
    	    	if (pos < otypes.size()) {
    	    		odef = (Odef) otypes.get(pos);
    	    		if (DEBUG) System.out.println("pos < size: odef="+odef.abbr+" otype&mask="+(odef.otype & mask));
    	    		if ( (odef.otype & mask) == o) {
    	    			return (true);
    	    		}
    	    	}
    	    	// Are there subtrees ? [0]=subtree [1]=class
    	    	while (inc < otrees.size()) {
    	    		odef = (Odef) otrees.get(inc);
    	    		if (DEBUG) System.out.println("subtrees: inc="+inc+" odef="+odef.abbr+" otype&omask="+(odef.otype & omask));
    	    		if ( (odef.otype & omask) == ori) {
    	    			break;
    	    		}
    	    		inc += 2;
    	    	}
    	    	if (inc >= otrees.size()) {
    	    		if (DEBUG) System.out.println("inc > otree.size");
    	    		return (false);
    	    	}
    	    	odef = (Odef) otrees.get(inc - 1);
    	    	pos = otypes.indexOf(odef);
    	    	o = odef.otype;
    	    	mask = Nlev[level(o)];
    	    	inc += 2;
    	    	if (DEBUG) System.out.println("fin true: odef="+odef.abbr+" pos="+pos+
 	    				" mask="+Integer.toHexString(mask)+" inc="+inc);
 	    		if (odef.abbr.compareTo("MoC") == 0 && nbmoc++ > 10) System.exit(0);
    	    	return (true);
    	    }

    	    public Object nextElement() {
    	    	Odef odef;
    	    	Otype e;
    	    	odef = (Odef) otypes.get(pos);
    	    	e = new Otype();
    	    	e.otype = odef.otype;
    	    	pos++;
    	    	return (e);
    	    }
    	};
    }

    //==================================================================
    //			Edition
    //==================================================================

    /** Default Edition of the Otype value
     * @param	choice 0=numeric, 1=standard, 2=verbose, 3=3-letter
     * @return	the meaning of the Object Type
     */
    public final String toString(int choice) {
    	Odef o;
    	int i;
    	if (! hasValue()) return ""; // NoData.NODATA_DISPLAY_STRING;
    	
    	if (choice == 0) {
    		StringBuffer b = new StringBuffer(40);
    		edit(b, otype);
    		return ("" + b);
    	}
    	o = new Odef(otype, null, null, null);
    	i = otypes.indexOf(o);
//    	if (i >= 0) {
	    	o = (Odef) otypes.get(i);
	    	if (choice == 1) {
	    		return (o.symb);
	    	}
	    	if (choice == 2) {
	    		return (o.expl);
	    	}
	    	if (choice == 3) {
	    		return (o.abbr);
	    	}
//    	} else {
    	return (null);
    }

    /** 
     * Default Edition of the Otype value
     * @return	the "standard" explanation of hte object type
     **/
    public final String toString() {
    	return (this.toString(1));
    }

    /**
     * Edition with options
     * @param opt Option string. It can contain:<br>
     * 'N'=numeric, 'S'=standard, 'V'=verbose, '3'=3-letter
     * and any other embedded character
     * @return the edited object type
     */
	  public String toString(String opt) {
	  	StringBuffer str = new StringBuffer();
		for (int k=0; k<opt.length(); k++) {
			char cc = opt.charAt(k);
			if (cc == '3' || cc == 'N' || cc == 'S' || cc == 'V') {
				str.append(toString(cc).trim());
			} else {
				str.append(cc);
			}
		}
	  	return str.toString();
	  }
//   public String toString(String opt) {
//		int p = "NSV3".indexOf(Character.toUpperCase(opt.charAt(0)));
//		return toString(p >= 0 ? p : 0);
//	  }
    
    /**
     * Edition with options
     * @param cc Option char. It can be:<br>
     * N=numeric, S=standard, V=verbose, 3=3-letter
     * @return the edited object
     */
    public String toString(char cc) {
    	switch (cc) {
    		case 'N':
    		default:
    			return toString(0);
    		case '3':
    			return toString(3);
    		case 'S':
    			return toString(1);
    		case 'V':
    			return toString(2);
    	}
    }
    
    /** 
     * Dump contents to stdout
     **/
    public final void dump() {
    	System.out.println("....Otype: " + Integer.toHexString(otype));
    }

    /**
     * Methode qui permet de transformer un otype decimal en hexadecimal
     * @return chaine representant la valeur de otype en hexadecimal
     */
    public String numericToHexa() {
    	StringBuffer str = new StringBuffer();
    	StringTokenizer st = new StringTokenizer(this.toString(0), ".");
    	while (st.hasMoreTokens()) {
    		String shex = Integer.toHexString(Integer.parseInt(st.nextToken()));
    		if (shex.length() == 1) {
    			shex = "0" + shex;
    		}
    		str.append(shex);
//			Parameters p = new Parameters();
//			str.append(Format.sprintf("%02x",p.add(Integer.parseInt(st.nextToken()))));
    	}
    	return str.toString();
    }

    /**
     * Methode qui transforme une chaine de caractere hexadecimale en otype sous forme numerique
     * @param strHex chaine representant un otype sous forme hexadecimale
     * @return otype sous forme numerique
     */
    public static String hexaToNumeric(String strHex) {
    	StringBuffer strtmp = new StringBuffer(strHex);
    	StringBuffer str = new StringBuffer();

    	if (strHex.compareTo("0") == 0) {
    		str.append("00.00.00.00");
    	}
    	else {
    		if (strHex.charAt(0) != '0') {
    			strtmp.insert(0, "0");
    		}
    		for (int i = 0; i < strtmp.length(); i += 2) {
    			str.append(Integer.parseInt(String.valueOf(strtmp.substring(i,
    				i + 2)), 16)).append("."); //Chaque caractere hexa est transforme en decimal
    		}
    		str.delete(str.length() - 1, str.length());
    	}

    	return str.toString();
    }

        /**
         * MOD-BB 21-01-04 Ajout de cette methode.
         * Test d'egalite de Otype.
         * @param o Objet a comparer.
         * @return Vrai si o est identique a this, faux sinon.
         */
        public boolean equals(Object o)
        {
          boolean res = false;
          if(o instanceof Otype)
            res = ((Otype)o).otype == this.otype;
          return res;
        }
        
    	/**
    	 * Check if the object type is unknown
    	 * @return true if it is unknown, false otherwise
    	 */
    	public boolean isUnknown() {
    		return otype == UNKNOWN;
    	}
    	
        /**
         * Copy values parameters
         * @return a copy of the object
         */
         public Object clone() {
    		try {
    			return super.clone();
    		} catch (CloneNotSupportedException e) {
    		// this shouldn't happen 
    		 	throw new InternalError();
    		 }
    	}

		/**
		 * Gives all explanation string of otypes
		 * @return array of first level strings
		 * @see #list()
		 */
		public String[] getList() {
			String[] s_otypes;
			s_otypes = new String[otypes.size()];

			for (int i = 0 ; i < otypes.size() ; i++)
				s_otypes[i] = ((Odef)otypes.get(i)).expl;
			
			return s_otypes;
		}

		/**
		 * @see cds.util.ValueList#getShortList()
		 */
		public String[] getShortList() {
			String[] s_otypes;

			s_otypes = new String[otypes.size()];

			for (int i = 0 ; i < otypes.size() ; i++)
				s_otypes[i] = ((Odef)otypes.get(i)).symb;
			
			return s_otypes;
		}

		/**
		 * Parse a string containing otype sting name
		 * @param s string to parse
		 * @return index of first char not used in the string
		 * @throws ParseException
		 * @see #set(java.lang.String)
		 */
		public int parse(String s) throws ParseException {
			return parse(s, 0);
		}

		/**
		 * Parse a string containing otype sting name
		 * @param s string to parse
		 * @param offset  first char to parse in the string
		 * @return index of first char not used in the string
		 * @throws ParseException
		 * @see #set(java.lang.String)
		 */
		public int parse(String s, int offset) throws ParseException {
			set(s.substring(offset));
			return s.length();
		}

		/**
		 * @return String object type
		 * 
		 */
		public String getName() {
			return toString(2);
		}
		
		/**
		 * @return String object type
		 */
		public String get() {
			return toString();
		}
		
		//==========================================================
		//       SQL queries utilities
		//==========================================================
		public int getQueryMask() {
			return getQueryMask(level(otype));
		}
		
		public int getQueryMask(int level) {
			return Nlev[level];
		}

		public int getQueryValue() {
			return otype;
		}
		
		
	    /**
	     * Compute the hashcode (overrides the Object method).
	     * The hashcode is the integer object type code
	     * @return the hashcode value for the otype
	     */
	    public int hashCode() {
	    	return otype & Nlev[0];
	    }


}

/**
 * Otype Definition object
 * This internal class defines the basic Otype symbols
 * loaded from a file.
 */
class Odef implements Serializable {
    public int otype; // The coded object type
    public String symb; // Main symbol of the Otype
    public String abbr; // 3-letter abbreviation of otype
    public String expl; // Long-text of the Symbol

   /**
    * Constructor of a otype object
    * @param o otype numeric value
    * @param s otype symbol (short name)
    * @param a otype abbreviation
    * @param text otype description
    */
    public Odef(int o, String s, String a, String text) {
    	//System.out.println("...Odef: " + s + " -- " + a + " -- " + text);
    	otype = o;
    	symb = s;
    	expl = text;
    	abbr = a;
    	if (a != null) {
    		int len = a.length();
    		if (len < 3) {
    			abbr = a + "   ".substring(0, 3 - len);
    		}
    	}
    }

   /**
    * Comparison of 2 Otypes
    * @param o Object for which we would like to compare the type
    * @return true if the two otypes are identical
   **/
    public final boolean equals(Object o) {
    	Odef odef = (Odef) o;
    	//System.out.println("....Compare " + this + " / " + odef);
    	return (otype == odef.otype);
    }
    
   /**
    * Display the object type in a string
    * @return the string containing the otype
    */
    public final String toString() {
    	StringBuffer b = new StringBuffer(80);
    	Otype.edit(b, otype);
    	b.append(": ");
    		if (abbr != null) {
    		b.append(abbr);
    		b.append(' ');
    		b.append('(');
    		b.append(expl);
    		b.append(')');
    	}
    	return ("" + b);
    }

   /**
    * Display for debugging purposes the otype on stdout
    */
    public final void dump() {
    	System.out.println("....Otypedef: " + toString());
    }
    
}

